#ifndef __mVec2_H__
#define __mVec2_H__

/*! 
    \file mVec2.h
    \brief 2D Vector
    \author Douglas H. Cox
*/

//============================================================================
//    REQUIRED HEADERS
//============================================================================
#include "mMath.h"

//============================================================================
//    DEFINITIONS / ENUMERATIONS / SIMPLE TYPEDEFS / CLASS FORWARDS
//============================================================================

//============================================================================
//    FUNCTION PROTOTYPES
//============================================================================

//============================================================================
//    CLASS PROTOTYPES
//============================================================================

//! 2D Vector / Point 
/*! mVec2 represents a vector consisting of an array of two float values. */
class MMATH_DLL mVec2
{
public:
  float v[2];   //!< vector data

  //! Default constructor
  /*! No initialization is done here. */
  mVec2() {}

  //! Copy constructor
  mVec2(const mVec2& src)
  { 
    v[0] = src.v[0]; 
    v[1] = src.v[1];
  }

  //! Constructor from 2 float values
  mVec2(float x, float y)
  {
    v[0] = x; 
    v[1] = y; 
  }

  //! Constructor from an array of 2 float values
  mVec2(float *src)
  {
    v[0] = src[0];
    v[1] = src[1];
  }

  //! Assignment
  void set(const mVec2& src)
  {
    v[0] = src.v[0]; 
    v[1] = src.v[1]; 
  }

  //! Assignment from 2 float values
  void set(float x, float y)
  {
    v[0] = x; 
    v[1] = y; 
  }

  //! Assignment from an arry of 2 float values
  void set(float *src)
  {
    v[0] = src[0];
    v[1] = src[1];
  }

  //! Assignment
  mVec2& operator = (const mVec2& src)
  { 
    v[0] = src.v[0]; 
    v[1] = src.v[1]; 
    return *this;
  }

  //! Accessor
  void get(float& x, float& y) const
  {
    x = v[0];
    y = v[1];
  }

  //! Accessor
  float& operator [] (int i)
  {
    return v[i];
  }

  //! Accessor
  float operator [] (int i) const
  {
    return v[i];
  }

  //! Returns true if this == a
  bool operator == (const mVec2& a) const
  {
    return( v[0] == a.v[0] &&
            v[1] == a.v[1] );
  }

  //! Returns true if this != a
  bool operator != (const mVec2& a) const
  { 
    return !(operator==(a));
  }

  //! Returns true is this is equal to a within some tolerance value.
  bool approxEqual(const mVec2& a, float tol = mZeroTol)
  {
    return( mApproxEqual(v[0], a.v[0], tol) &&
            mApproxEqual(v[1], a.v[1], tol) );
  }


  //! Sets each component to the min of this and a.
  void setMin(const mVec2& a)
  {
    if (a.v[0] < v[0]) v[0] = a.v[0];
    if (a.v[1] < v[1]) v[1] = a.v[1];
  }

  //! Sets each component to the max of this and a.
  void setMax(const mVec2& a)
  {
    if (a.v[0] > v[0]) v[0] = a.v[0];
    if (a.v[1] > v[1]) v[1] = a.v[1];
  }

  //! Clamps each component of this to [a, b]
  void clamp(float a, float b)
  {
    v[0] = mClamp(v[0], a, b);
    v[1] = mClamp(v[1], a, b);
  }

  //! Clamps each component of vec to [a, b]
  void clamp(const mVec2& vec, float a, float b)
  {
    v[0] = mClamp(vec.v[0], a, b);
    v[1] = mClamp(vec.v[1], a, b);
  }

  //! Addition: this += a
  void add(const mVec2& a)
  {
    v[0] += a.v[0];
    v[1] += a.v[1];
  }

  //! Addition: this = a + b
  void add(const mVec2& a, const mVec2& b)
  {
    v[0] = a.v[0] + b.v[0];
    v[1] = a.v[1] + b.v[1];
  }

  //! Addition: Returns this + a
  mVec2 operator + (const mVec2& a) const
  {
    return mVec2(v[0] + a.v[0],
                 v[1] + a.v[1]);
  }

  //! Addition: this += a
  mVec2& operator += (const mVec2& a)
  {
    v[0] += a.v[0];
    v[1] += a.v[1];
    return *this;
  }

  //! this += a * s
  void addScaled(const mVec2& a, float s)
  {
    v[0] += a.v[0] * s;
    v[1] += a.v[1] * s;
  }

  //! this = a + b * s
  void addScaled(const mVec2& a, const mVec2& b, float s)
  {
    v[0] = a.v[0] + b.v[0] * s;
    v[1] = a.v[1] + b.v[1] * s;
  }

  //! this = a * s1 + b * s2
  void addScaled(const mVec2& a, float s1,
                 const mVec2& b, float s2)
  {
    v[0] = a.v[0] * s1 + b.v[0] * s2;
    v[1] = a.v[1] * s1 + b.v[1] * s2;
  }

  //! Subtraction: this -= a
  void sub(const mVec2& a)
  {
    v[0] -= a.v[0];
    v[1] -= a.v[1];
  }

  //! Subtraction: this = a - b
  void sub(const mVec2& a, const mVec2& b)
  {
    v[0] = a.v[0] - b.v[0];
    v[1] = a.v[1] - b.v[1];
  }

  //! Subtraction: Returns this - a
  mVec2 operator - (const mVec2& a) const
  {
    return mVec2(v[0] - a.v[0],
                 v[1] - a.v[1]);
  }

  //! Subtraction: this -= a
  mVec2& operator -= (const mVec2& a)
  {
    v[0] -= a.v[0];
    v[1] -= a.v[1];
    return *this;
  }

  //! Scales each component by s.
  void scale(float s)
  {
    v[0] *= s; 
    v[1] *= s; 
  }

  //! Scales each component of a by s.
  void scale(const mVec2& a, float s)
  {
    v[0] = a.v[0] * s;
    v[1] = a.v[1] * s;
  }

  //! Scales each component by s.
  mVec2& operator *= (float s)
  {
    v[0] *= s; 
    v[1] *= s;
    return *this;
  }

  //! Negate: this = -this
  void neg()
  {
    v[0] = -v[0];
    v[1] = -v[1];
  }

  //! Negate: this = -a
  void neg(const mVec2& a)
  {
    v[0] = -a.v[0];
    v[1] = -a.v[1];
  }

  //! Negate: Returns -this
  mVec2 operator - () const
  {
    return mVec2(-v[0], -v[1]);
  }

  //! Component multiplication
  void compMult(const mVec2& a)
  {
    v[0] *= a[0];
    v[1] *= a[1];
  }

  //! Component multiplication
  void compMult(const mVec2& a, const mVec2& b)
  {
    v[0] = a[0] * b[0];
    v[1] = a[1] * b[1];
  }

  //! Component division
  void compDiv(const mVec2& a)
  {
    v[0] /= a[0];
    v[1] /= a[1];
  }

  //! Component division
  void compDiv(const mVec2& a, const mVec2& b)
  {
    v[0] = a[0] * b[0];
    v[1] = a[1] * b[1];
  }

  //! Returns the distance squared between this and a.
  float distance2(const mVec2& a) const
  {
    return( mSquare(v[0] - a.v[0]) +
            mSquare(v[1] - a.v[1]) );
  }

  //! Returns the distance between this and a.
  float distance(const mVec2& a) const
  {
    return( mSqrt(distance2(a)) );
  }

  //! Returns the length squared of this.
  float length2() const
  {
    return( mSquare(v[0]) + 
            mSquare(v[1]) );
  }

  //! Returns the length of this.
  float length() const
  {
    return( mSqrt(length2()) );
  }

  //! Returns the dot product of this and a.
  float dot(const mVec2& a) const
  {
    return( v[0] * a.v[0] +
            v[1] * a.v[1] );
  }

  //! Normalizes this and returns the previous length.
  float normalize()
  {
    float l(length());
    if (l != 0.0f) scale(1.0f / l);
    return l;
  }

  //! Linearly interpolates between two vectors.
  void lerp(const mVec2& a, const mVec2& b, float t)
  {
    v[0] = a.v[0] + t * (b.v[0] - a.v[0]);
    v[1] = a.v[1] + t * (b.v[1] - a.v[1]);
  }

};

// Common constant values.
const mVec3 mZero2(0.0f, 0.0f);
const mVec3 mOne2(1.0f, 1.0f);

//****************************************************************************
//**    END HEADER mVec2.h
//****************************************************************************
#endif // __mVec2_H__
