//****************************************************************************
//**
//**    mFrame.cpp
//**    Implementation for mFrame.
//**    (C) 2000 - Douglas H. Cox
//**
//****************************************************************************
//============================================================================
//    HEADERS
//============================================================================
#include "mFrame.h"

#include "mQuat.h"

//============================================================================
//    PRIVATE DEFINITIONS / ENUMERATIONS / SIMPLE TYPEDEFS
//============================================================================

//============================================================================
//    PRIVATE STRUCTURES / UTILITY CLASSES
//============================================================================

//============================================================================
//    PRIVATE DATA
//============================================================================

//============================================================================
//    PRIVATE FUNCTIONS
//============================================================================

//============================================================================
//    FUNCTIONS
//============================================================================

//============================================================================
//    CLASS METHODS
//============================================================================

//=============================================================
// mFrame::approxEqual
bool mFrame::approxEqual(const mFrame& f, float tol) const
{
  return (a[0].approxEqual(f.a[0], tol) &&
          a[1].approxEqual(f.a[1], tol) &&
          a[2].approxEqual(f.a[2], tol));          
}

//=============================================================
// mFrame::inverse
void mFrame::inverse()
{
  float tmp;
  tmp = a[0].v[1];
  a[0].v[1] = a[1].v[0];
  a[1].v[0] = tmp;

  tmp = a[0].v[2];
  a[0].v[2] = a[2].v[0];
  a[2].v[0] = tmp;

  tmp = a[1].v[2];
  a[1].v[2] = a[2].v[1];
  a[2].v[1] = tmp;
}

//=============================================================
// mFrame::inverse
void mFrame::inverse(const mFrame& f)
{
  float tmp;
  tmp = f.a[0].v[1];
  a[0].v[1] = f.a[1].v[0];
  a[1].v[0] = tmp;

  tmp = f.a[0].v[2];
  a[0].v[2] = f.a[2].v[0];
  a[2].v[0] = tmp;

  tmp = f.a[1].v[2];
  a[1].v[2] = f.a[2].v[1];
  a[2].v[1] = tmp;
  
  // diagonal
  a[0].v[0] = f.a[0].v[0];
  a[1].v[1] = f.a[1].v[1];
  a[2].v[2] = f.a[2].v[2];
}

//=============================================================
// mFrame::xformTo
void mFrame::xformTo(const mFrame& f, const mFrame& to)
{
  mFrame nf;

  nf.a[0].xformTo(f.a[0], to);
  nf.a[1].xformTo(f.a[1], to);
  nf.a[2].xformTo(f.a[2], to);

  set(nf);
}

//=============================================================
// mFrame::xformFrom
void mFrame::xformFrom(const mFrame& f, const mFrame& from)
{
  mFrame nf;

  nf.a[0].xformFrom(f.a[0], from);
  nf.a[1].xformFrom(f.a[1], from);
  nf.a[2].xformFrom(f.a[2], from);

  set(nf);
}
 
//=============================================================
// mFrame::setQuat
void mFrame::setQuat(const mQuat& q)
{
  // standard quaternion to matrix operations
  float x(q.v[0]), y(q.v[1]), z(q.v[2]), w(q.v[3]);
  float x2(x*2.0f), y2(y*2.0f), z2(z*2.0f), w2(w*2.0f);
	float xx2(x*x2), yy2(y*y2), zz2(z*z2), ww2(w*w2);
	float xy2(x*y2), xz2(x*z2), xw2(x*w2), yz2(y*z2), yw2(y*w2), zw2(z*w2);

  a[0].set(1.0f - (yy2 + zz2), xy2 - zw2, xz2 + yw2);
  a[1].set(xy2 + zw2, 1.0f - (xx2 + zz2), yz2 - xw2);
  a[2].set(xz2 - yw2, yz2 + xw2, 1.0f - (xx2 + yy2));
}

//=============================================================
// mFrame::getQuat
void mFrame::getQuat(mQuat& q) const
{
  float tr, s;
  tr = a[0].v[0] + a[1].v[1] + a[2].v[2];

  if (tr > 0.0f)
  {
    s = mSqrt(tr + 1.0f);
    q.v[3] = s*0.5f;
    s = 0.5f / s;

    q.v[0] = s * (a[2].v[1] - a[1].v[2]);
    q.v[1] = s * (a[0].v[2] - a[2].v[0]);
    q.v[2] = s * (a[1].v[0] - a[0].v[1]);
  }
  else
  {
    const int n[3] = {1, 2, 0};
    int i, j, k;

    i = 0;
    if (a[1].v[1] > a[0].v[0])
      i = 1;
    if (a[2].v[2] > a[i].v[i])
      i = 2;
    j = n[i];
    k = n[j];

    s = mSqrt((a[i].v[i] - (a[j].v[j] + a[k].v[k])) + 1.0f);

    q.v[i] = s * 0.5f;
    assert(s !=  0.0f);

    s = 0.5f / s;

    q.v[3] = (a[k].v[j] - a[j].v[k]) * s;
    q.v[j] = (a[j].v[i] + a[i].v[j]) * s;
    q.v[k] = (a[k].v[i] + a[i].v[k]) * s;
    
    // TODO: Test this.
    bool bNotYetTested = false;
    assert(bNotYetTested);
  }
}

//=============================================================
// mFrame::setAxisAngle
void mFrame::setAxisAngle(const mVec3& n, float a)
{
  mQuat q;
  q.setAxisAngle(n, a);
  setQuat(q);
}

//=============================================================
// mFrame::getAxisAngle
void mFrame::getAxisAngle(mVec3& n, float& a) const
{
  mQuat q;
  getQuat(q);
  q.getAxisAngle(n, a);
}

// ============================================================
// mFrame::setHP
void mFrame::setHP(float heading, float pitch)
{
  float sh, ch, sp, cp;
  mSinCos(heading, sh, ch);
  mSinCos(pitch, sp, cp);

  a[0].set(ch, sh, 0.0f);
  a[1].set(-cp*sh, ch*cp, sp);
  a[2].set(sp*sh, sp*ch, cp);
}

// ============================================================
// mFrame::setHPR
void mFrame::setHPR(float heading, float pitch, float roll)
{
  float sh, ch, sp, cp, sr, cr;
  mSinCos(heading, sh, ch);
  mSinCos(pitch, sp, cp);
  mSinCos(roll, sr, cr);

  float srsp = sr*sp;
  float crsp = cr*sp;

  a[0].set(cr*ch - srsp*sh, cr*sh + srsp*ch, -sr * cp);
  a[1].set(-cp*sh, ch*cp, sp);
  a[2].set(sr*ch + crsp*sh, sr*sh - crsp*ch, cr * cp);
}


//****************************************************************************
//**    END MODULE mFrame.cpp
//****************************************************************************

