//****************************************************************************
//**
//**    mQuat.cpp
//**    Implementation for mQuat.
//**    (C) 2000 - Douglas H. Cox
//**
//****************************************************************************
//============================================================================
//    HEADERS
//============================================================================
#include "mQuat.h"
#include "mVec3.h"

//============================================================================
//    PRIVATE DEFINITIONS / ENUMERATIONS / SIMPLE TYPEDEFS
//============================================================================

//============================================================================
//    PRIVATE STRUCTURES / UTILITY CLASSES
//============================================================================

//============================================================================
//    PRIVATE DATA
//============================================================================

//============================================================================
//    PRIVATE FUNCTIONS
//============================================================================

//============================================================================
//    FUNCTIONS
//============================================================================

//============================================================================
//    CLASS METHODS
//============================================================================

//=============================================================
// mQuat::inverse
void mQuat::inverse()
{
  // q.scale(1.0f/length2()) 
  // q.conjugate()
  scale(-1.0f / length2());
  v[3] = -v[3];
}

//=============================================================
// mQuat::inverse
void mQuat::inverse(const mQuat& q)
{
  // q.scale(1.0f/length2()) 
  // q.conjugate()
  scale(q, -1.0f / q.length2());
  v[3] = -v[3];
}

// ============================================================
// mQuat::mult
void mQuat::mult(const mQuat& q1, const mQuat& q2)
{
  mQuat tmp;
  tmp[0] = q1[3] * q2[0] + q1[0] * q2[3] +
           q1[1] * q2[2] - q1[2] * q2[1];
  tmp[1] = q1[3] * q2[1] + q1[1] * q2[3] +
           q1[2] * q2[0] - q1[0] * q2[2];
  tmp[2] = q1[3] * q2[2] + q1[2] * q2[3] +
           q1[0] * q2[1] - q1[1] * q2[0];
  tmp[3] = q1[3] * q2[3] - q1[0] * q2[0] -
           q1[1] * q2[1] - q1[2] * q2[2];
  set(tmp);
}

//=============================================================
// mQuat::setAxisAngle
void mQuat::setAxisAngle(const mVec3& n, float a)
{
  float s, c;
  mSinCos(a * 0.5f, s, c);

  v[0] = n[0] * s;
  v[1] = n[1] * s;
  v[2] = n[2] * s;
  v[3] = c;
}

//=============================================================
// mQuat::getAxisAngle
void mQuat::getAxisAngle(mVec3& n, float& a) const
{
  float s = 1.0f / mSqrt( 1.0f - v[3]*v[3] );

  a = mACos(v[3]) * 2.0f;
  n[0] = v[0] * s;
  n[1] = v[1] * s;
  n[2] = v[2] * s; 
}

// ============================================================
// mQuat::setHP
void mQuat::setHP(float h, float p)
{
  // TODO: check h/p > 180.0f etc.

  h *= 0.5f; 
  p *= 0.5f;

  float cH, sH;
  float cP, sP;

  mSinCos(h, sH, cH);
  mSinCos(p, sP, cP);

  v[0] = -(sP * cH);
  v[1] = -(sP * sH);
  v[2] = -(cP * sH);
  v[3] =  (cP * cH);
}

//=============================================================
// mQuat::getHP
void mQuat::getHP(float &h, float &p) const
{
  h = 2.0f * mATan2(-v[2], v[3]);
  p = 2.0f * mATan2(-v[0], v[3]);
}

// ============================================================
// mQuat::setHPR
void mQuat::setHPR(float h, float p, float r)
{
  h *= 0.5f; 
  p *= 0.5f;
  r *= 0.5f;

  float cH, sH;
  float cP, sP;
  float cR, sR;

  mSinCos(h, sH, cH);
  mSinCos(p, sP, cP);
  mSinCos(r, sR, cR);

  float cPcH = cP * cH;
  float cPsH = cP * sH;
  float sPcH = sP * cH;
  float sPsH = sP * sH;

  v[0] = -(sPcH * cR) + (cPsH * sR);
  v[1] = -(cPcH * sR) - (sPsH * cR);
  v[2] = -(cPsH * cR) - (sPcH * sR);
  v[3] =  (cPcH * cR) - (sPsH * sR);
}

//=============================================================
// mQuat::getHPR
void mQuat::getHPR(mVec3& hpr) const
{
  assert(false); // TODO: implement
}

//=============================================================
// mQuat::slerp
void mQuat::slerp(const mQuat& q1, const mQuat& q2, float t)
{
  float cosom = q1.dot(q2);
  float s1, s2;

  if ( (1.0f + cosom) > mZeroTol)
  {
    if ( (1.0f - cosom) > mZeroTol)
    {
      double omega, isinom;
      omega = acos(cosom);
      isinom = 1.0f / sin(omega);
      s1 = float(sin((1.0f - t) * omega) * isinom);
      s2 = float(sin(t * omega) * isinom);
    }
    else
    {
      s1 = 1.0f - t;
      s2 = t;
    }
    addScaled(q1, s1, q2, s2);
  }
  else
  {
    mQuat pq;
    pq.v[0] = -q2.v[1];
    pq.v[1] = q2.v[0];
    pq.v[2] = -q2.v[3];
    pq.v[3] = q2.v[2];

    s1 = float(sin((1.0f - t) * mHalfPI));
    s2 = float(sin(t * mHalfPI));
    addScaled(q1, s1, pq, s2);
  }
}

//****************************************************************************
//**    END MODULE mQuat.cpp
//****************************************************************************

